import numpy as np

from poker_game_runner.InfoState import Observation

def get_name():    
    return "randomBot"

def act(observation: Observation):
    return np.random.choice(observation.legal_actions)


from dataclasses import dataclass
from typing import List, Sequence, Tuple

RANKS = '23456789TJQKA'
SUITS = 'cdhs'

@dataclass(frozen=True)
class PlayerInfo: 
    spent: int
    stack: int
    active: bool

@dataclass(frozen=True)
class ActionInfo: 
    player: int
    action: int
    def __str__(self) -> str:
        if self.player == -1:
            return "Dealer: " + card_num_to_str(self.action)
        player_str = "Player: " + str(self.player)
        if self.action == 0:
            return player_str + "Fold"
        if self.action == 1:
            return player_str + "Call"
        return player_str + "Raise to " + str(self.action)

@dataclass(frozen=True)
class Observation: 
    my_hand: Tuple[str]
    my_index: int
    board_cards: Tuple[str]
    player_infos: Tuple[PlayerInfo]
    history: Tuple[ActionInfo]
    small_blind: int
    big_blind: int
    legal_actions: Tuple[int]

    def get_player_count(self):
        return len(self.player_infos)

    def get_active_player_count(self):
        return len(p for p in self.player_infos if p.active)

    def get_betting_round(self):
        deal_count = len(h for h in self.history if h.player == -1)
        return 0 if deal_count == 0 else deal_count-2

    def get_max_spent(self):
        return max(map(lambda p: p.spent, self.player_infos))

    def get_asked_amount(self):
        return self.get_max_spent()
    
    def can_raise(self):
        return any(a for a in self.legal_actions if a > 1)
    
    def get_min_raise(self):
        return min(a for a in self.legal_actions if a > 1) if self.can_raise() else 0

    def get_max_raise(self):
        return max(a for a in self.legal_actions if a > 1) if self.can_raise() else 0

    def action_to_str(self, action_num: int, player_idx: int = None):
        if player_idx is None:
            player_idx = self.my_index
        return str(ActionInfo(player_idx, action_num))

class InfoState: 
    player_hands: List[Tuple[str]]
    board_cards: List[str]
    player_infos: List[PlayerInfo]
    history: List[ActionInfo]
    small_blind: int
    big_blind: int

    def __init__(self, history: List[int], stacks: List[int], blinds: List[int]):
        cards1 = map(card_num_to_str, [history[i] for i in range(0,len(history),2)])
        cards2 = map(card_num_to_str, [history[i] for i in range(1,len(history),2)])
        self.player_hands = [hand for hand in zip(cards1, cards2)]
        self.player_infos = [PlayerInfo(blind, stack-blind, True) for blind, stack in zip(blinds, stacks)]
        self.board_cards = []
        self.history = []
        self.small_blind = blinds[0]
        self.big_blind = blinds[1]

    def update_info_state_action(self, player_idx: int, action: int):
        player_info = self.player_infos[player_idx]
        if action > 1:
            r = action - player_info.spent
            self.player_infos[player_idx] = PlayerInfo(action, player_info.stack - r, player_info.active)
        elif action == 1:
            max_spent = max(map(lambda p: p.spent, self.player_infos))
            c = max_spent - player_info.spent
            self.player_infos[player_idx] = PlayerInfo(max_spent, player_info.stack - c, player_info.active)
        else:
            self.player_infos[player_idx] = PlayerInfo(player_info.spent, player_info.stack, False)
        self.history.append(ActionInfo(player_idx, action))


    def update_info_state_draw(self, card_num = int):
        cards_str = card_num_to_str(card_num)
        self.board_cards.append(cards_str)
        self.history.append(ActionInfo(-1,card_num))
            
    def to_observation(self, player_idx: int, legal_actions: List[int]):
        return Observation(self.player_hands[player_idx], 
                        player_idx, 
                        tuple(self.board_cards), 
                        tuple(self.player_infos), 
                        tuple(self.history),
                        self.small_blind,
                        self.big_blind,
                        tuple(legal_actions))

def amount_to_call(player_infos: Sequence[PlayerInfo], player_idx: int):
    max_spent = max(map(lambda p: p.spent, player_infos))
    return max_spent - player_infos[player_idx].spent

def card_num_to_str(card_num: int):
    rank_num = int(card_num / 4)
    suit_num = card_num % 4
    return RANKS[rank_num] + SUITS[suit_num]


import numpy as np
import pyspiel
from poker_game_runner import InfoState
from typing import List, Tuple
from collections import namedtuple

BlindScheduleElement = namedtuple('BlindScheduleElement', 'next_blind_change small_blind big_blind ante')
Player = namedtuple('Player', 'bot_impl stack')

def play_hand(players: List[Player], blinds: List[int]):

    game = pyspiel.load_game("universal_poker", {
        "betting": "nolimit",
        "bettingAbstraction": "fullgame",
        "numPlayers": len(players),
        "stack": " ".join(str(player.stack) for player in players),
        "blind": " ".join(str(blind) for blind in blinds),
        "numRounds": 4,
        "numHoleCards": 2,
        "numBoardCards": "0 3 1 1",
        "numSuits": 4,
        "numRanks": 13,
        "firstPlayer": "3 1 1 1" if len(players) > 2 else "1 1 1 1"
    })
    state = game.new_initial_state()

    #deal private cards
    while state.is_chance_node():
        state.apply_action(np.random.choice(state.legal_actions()))
        continue

    info_state = InfoState.InfoState(state.history(), [p.stack for p in players], [b for b in blinds])

    while not state.is_terminal():
        if state.is_chance_node():
            card_num = np.random.choice(state.legal_actions())
            state.apply_action(card_num)
            info_state.update_info_state_draw(card_num)
            continue
        
        current_idx = state.current_player()
        observation = info_state.to_observation(current_idx, state.legal_actions())
        action = players[current_idx].bot_impl.act(observation)
        if not action in state.legal_actions():
            if 0 in state.legal_actions():
                action = 0
            else:
                action = 1
        state.apply_action(action)
        info_state.update_info_state_action(current_idx, action)

    return map(int, state.rewards())

def play_tournament_table(bots, start_stack: int, blind_schedule: Tuple[BlindScheduleElement]):

    active_players = [Player(bot,start_stack) for bot in bots]
    np.random.shuffle(active_players)

    results = []
    hand_count = 0
    blinds_iter = iter(blind_schedule)
    current_blinds = next(blinds_iter)
    if current_blinds.big_blind > start_stack:
        print("big_blind cannot be bigger than start_stack")
        return []

    while len(active_players) > 1:
        print(sorted([player.bot_impl.get_name() for player in active_players]))

        rewards = play_hand(active_players, get_blinds_input(current_blinds, len(active_players)))

        if hand_count == current_blinds.next_blind_change:
            current_blinds = next(blinds_iter)

        defeated_players, active_players = update_active_players(active_players, rewards, current_blinds.big_blind)

        results = results + defeated_players
        active_players = active_players[1:] + [active_players[0]]
        hand_count += 1
    
    results = results + [active_players[0].bot_impl.get_name()]
    results.reverse()
    return results

def update_active_players(active_players: List[Player], rewards: List[int], big_blind: int):    
    updated_players = [Player(player.bot_impl, int(player.stack+r)) for player,r in zip(active_players, rewards)]

    defeated_players = [player.bot_impl.get_name() for player in updated_players if player.stack < big_blind]
    active_players = [player for player in updated_players if player.stack >= big_blind]
    return defeated_players, active_players

def get_blinds_input(current_blinds: BlindScheduleElement, playerCount: int) -> List[int]:
    return [current_blinds.small_blind, current_blinds.big_blind] + ([current_blinds.ante] * (playerCount-2))